use log::debug;
use std::fs::File;
use std::io::prelude::*;
use std::path::Path;

use anyhow::{anyhow, Result};
use async_trait::async_trait;
use base64;
use bytes::Bytes;
use futures::future::{JoinAll, OptionFuture};
use html_minifier::HTMLMinifier;
use indoc::{formatdoc, indoc};
use itertools::izip;
use nipper::Document;
use url::Url;

use crate::consts::EVENT_HANDLERS;
use crate::errors;
use crate::Download;

/// Makes an html page self-contained
///
/// The `downloader` must implement `Download` and is used to download ressources that are
/// needed to make this page self-contained such as stylesheets or images.
///
/// The function also removes all scripts on the page
pub async fn self_contained_html<E, D, S>(html: S, downloader: &D, base_url: &Url) -> String
where
    E: std::error::Error,
    D: Download<Error = E> + Send,
    S: AsRef<str>,
{
    // TODO: split/refactor this function
    //       ¿ find a way to ease the |get urls| -> |async download| -> |replace| workflow ?
    let (style_urls, html) = {
        let document = Document::from(html.as_ref());

        // ---- Remove  scripts ----
        //
        document.select("script").remove();

        for event in EVENT_HANDLERS {
            document
                .select(format!("[{}]", event).as_str())
                .remove_attr(event);
        }

        // ---- Replace stylesheets ----
        //
        let stylesheets = document.select("link[href][rel=\"stylesheet\"]");
        let styles_url = stylesheets
            .iter()
            .map(|stylesheet| {
                if let Some(src) = stylesheet.attr("href") {
                    //TODO: does it work with absolute urls ?
                    base_url.join(src.as_ref()).ok()
                } else {
                    None
                }
            })
            .collect::<Vec<_>>();
        (styles_url, String::from(document.html()))
    };

    let style_urls = style_urls.into_iter().map(|style_url| {
        OptionFuture::from(style_url.map(|s| async move { downloader.download(&s).await.unwrap() }))
    });
    let downloaded_styles = futures::future::join_all(style_urls).await;

    let html = {
        let document = Document::from(&html);
        let styles = document.select("link[href][rel=\"stylesheet\"]");

        styles
            .iter()
            .zip(downloaded_styles.iter())
            .for_each(|(mut stylesheet, inner_css)| {
                if let Some(inner_css) = inner_css {
                    let css = String::from_utf8(inner_css.to_vec()).unwrap();
                    let css = format!("<style>{}</style>", css);
                    stylesheet.replace_with_html(css);
                } else {
                    stylesheet.remove();
                }
            });
        String::from(document.html())
    };

    // ---- Replace imgs ----
    //
    let image_urls = {
        let document = Document::from(&html);
        let imgs = document.select("img");

        imgs.iter()
            .map(|image| {
                if let Some(src) = image.attr("src") {
                    base_url.join(src.as_ref()).ok()
                } else {
                    None
                }
            })
            .collect::<Vec<_>>()
    };

    let downloaded_images = image_urls.into_iter().map(|image_url| {
        OptionFuture::from(image_url.map(|url| async move {
            let data = downloader.download(&url).await.unwrap();
            (url, data)
        }))
    });
    let downloaded_images = futures::future::join_all(downloaded_images).await;

    let html = {
        let document = Document::from(&html);
        let imgs = document.select("img");

        imgs.iter()
            .zip(downloaded_images.iter())
            .for_each(|(mut img, data)| {
                if let Some((url, data)) = data {
                    let data = base64::encode(data);
                    let extension = Path::new(url.path()).extension().unwrap().to_str().unwrap();
                    img.set_attr("src", &format!("data:image/{};base64,{}", extension, data));
                }
            });
        String::from(document.html())
    };

    let mut minifier = HTMLMinifier::new();
    minifier.digest(html.as_str()).unwrap();

    String::from_utf8(minifier.get_html().into()).unwrap()
}

#[cfg(test)]
mod tests {
    use super::*;

    fn init() {
        let _ = env_logger::builder().is_test(true).try_init();
    }

    // TODO: the Dummy,Css and Png Downloaders don't really test the async scenario as
    //       they don't use futures : they don't call await.
    //       They should be testing the async scenario
    struct DummyDownloader;
    #[async_trait]
    impl Download for DummyDownloader {
        type Error = errors::Error;
        async fn download(&self, _file_link: &Url) -> errors::Result<Bytes> {
            Ok(Bytes::from(""))
        }
    }

    #[tokio::test]
    async fn remove_scripts() -> Result<()> {
        let html = "<html><head><script>let id = id => id</script></head><body></body></html>";
        let base_url = Url::parse("http://example.com")?;
        let downloader = DummyDownloader {};
        assert_eq!(
            self_contained_html(html, &downloader, &base_url).await,
            "<html><head></head><body></body></html>"
        );
        Ok(())
    }

    #[tokio::test]
    async fn remove_onevent_handlers() -> Result<()> {
        init();
        let downloader = DummyDownloader {};
        let html = |onevent| {
            formatdoc! {"
                <html>
                    <head>
                    </head>
                    <body>
                        <button class=\"activate\" {}=\"let id = id => id\">button</button>
                    </body>
                </html>",
                onevent
            }
        };

        let base_url = Url::parse("http://example.com")?;
        for s in EVENT_HANDLERS {
            assert_eq!(
                self_contained_html(html(s), &downloader, &base_url).await,
                "<html><head>\n</head>\n<body>\n<button class=\"activate\">button</button>\n</body></html>"
            );
        }
        Ok(())
    }

    struct CssDownloader;
    #[async_trait]
    impl Download for CssDownloader {
        type Error = errors::Error;
        async fn download(&self, _file_link: &Url) -> errors::Result<Bytes> {
            Ok(indoc! {"
                section#warning {
                    color: red;
                }"}
            .into())
        }
    }

    #[tokio::test]
    async fn download_css() -> Result<()> {
        let downloader = CssDownloader {};

        let html = indoc! {"
            <html>
                <head>
                    <link rel=\"stylesheet\" href=\"main.css\">
                </head>
                <body>
                </body>
            </html>
        "};

        // FIXME: find why minify doesn't minify
        let wanted_html = indoc! {"
            <html><head>
                    <style>
                        section#warning {
                            color: red;
                        }
                    </style>
                </head>
                <body>
                </body></html>
        "};
        let mut minifier = HTMLMinifier::new();
        minifier.digest(wanted_html)?;
        let minified = String::from_utf8(minifier.get_html().into())?;

        let base_url = Url::parse("http://example.com")?;
        assert_eq!(
            self_contained_html(html, &downloader, &base_url).await,
            minified
        );
        Ok(())
    }

    struct PngDownloader;
    #[async_trait]
    impl Download for PngDownloader {
        type Error = errors::Error;
        async fn download(&self, _file_link: &Url) -> errors::Result<Bytes> {
            let image_path = Path::new("test_data/home.png");
            let mut image_file = File::open(&image_path).unwrap();
            let mut image_buf: Vec<u8> = vec![];
            image_file.read_to_end(&mut image_buf).unwrap();
            Ok(image_buf.into())
        }
    }

    #[tokio::test]
    async fn download_image_png() -> Result<()> {
        let downloader = PngDownloader {};
        let html = indoc! {"
            <html>
                <head></head>
                <body>
                    <img src=\"home.png\" alt=\"an home\" />
                </body>
            </html>
        "};

        let wanted_html = indoc! {"
            <html><head></head>
                <body>
                <img src=\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABm\
                JLR0QA/wD/AP+gvaeTAAAAh0lEQVQ4jc2RQQqAIBBFn12ioNNE544gWhStc+FNWtVmBDFHsjZ9+DBffeOgo\
                KsBVmCRukgNYIFTbGVtlDzk4DqCwyZhTqoFXAJO+RN8a1ADewF8CvPqZm8nLNsL2HutgEN70Qc6TBDUr1Fk\
                AKrMgU4OGaDPdlEmMFFO7ucmeKR/NZgLuMkXFxHZVhLI8sXeAAAAAElFTkSuQmCC\" alt=\"an home\">
                </body></html>
        "};
        let mut minifier = HTMLMinifier::new();
        minifier.digest(wanted_html)?;
        let minified = String::from_utf8(minifier.get_html().into())?;

        let base_url = Url::parse("http://example.com")?;
        assert_eq!(
            self_contained_html(html, &downloader, &base_url).await,
            minified
        );
        Ok(())
    }
}
